import logging


try:
    from . import hollow_controller as hollow_controller
except Exception:
    import n0s1.controllers.hollow_controller as hollow_controller


class GitHubController(hollow_controller.HollowController):
    def __init__(self):
        super().__init__()
        self._client = None

    def set_config(self, config=None):
        super().set_config(config)
        from github import Github
        TOKEN = config.get("token", "")
        self._client = Github(TOKEN)
        self._owner = config.get("owner", "")
        self._repo = config.get("repo", "")
        self._branch = config.get("branch", "")
        return self.is_connected()

    def get_name(self):
        return "GitHub"

    def is_connected(self):
        if self._client:
            if user := self._client.get_user():
                self.log_message(f"Logged to {self.get_name()} as {user}")
                return True
            else:
                self.log_message(f"Unable to connect to {self.get_name()}. Check your credentials.", logging.ERROR)
                return False
        return False

    def _get_repo_fullname(self, repo_name):
        owner = self._owner
        repo_fullname = ""

        git_suffix = ".git"
        suffix_len = len(git_suffix)
        repo_name_len = len(repo_name)
        git_url_suffix = repo_name.lower()[repo_name_len-suffix_len:]
        if git_url_suffix == git_suffix:
            repo_fullname = repo_name.lower().replace("git://github.com/", "")
            repo_fullname = repo_fullname[:-suffix_len]
            return repo_fullname

        if not owner or len(owner) <= 0:
            owner = self._client.get_user().login
        repo_fullname = owner + "/" + repo_name
        return repo_fullname

    def _get_repo_obj(self, repo_gid=None):
        if self._repo and len(self._repo) > 0:
            repo_name = self._get_repo_fullname(self._repo)
            return self._client.get_repo(repo_name)
        full_name = self._get_repo_fullname(repo_gid)
        return self._client.get_repo(full_name)

    def _get_repos(self):
        owner = {}
        repos = []

        if self._scan_scope:
            owner = self._scan_scope.get("owner", {})
            for key in self._scan_scope.get("repos", {}):
                repos.append(key)
        if len(repos) > 0:
            return repos, owner
        self.connect()
        if self._owner and len(self._owner) > 0:
            from github import UnknownObjectException
            try:
                org = self._client.get_organization(self._owner)
                repos = org.get_repos()
                owner = {"type": "org", "name": self._owner}
            except UnknownObjectException as e:
                try:
                    message = f"Unable to get ORG {self._owner} as owner: {e}"
                    self.log_message(message, logging.WARNING)
                    message = f"Trying to get user {self._owner} as owner..."
                    self.log_message(message, logging.WARNING)
                    repos = self._client.get_user(self._owner).get_repos()
                    owner = {"type": "user", "name": self._owner}
                except Exception as e:
                    message = f"Unable to get repos from {self._owner}: {e}"
                    self.log_message(message, logging.ERROR)
        else:
            user = self._client.get_user()
            repos = user.get_repos()
            owner = {"type": "authenticated_user", "name": user.login}

        if self._repo and len(self._repo) > 0:
            for r in repos:
                if r.name.lower() == self._repo.lower():
                    return [r], owner
            return [], owner

        return repos, owner

    def _filter_branches(self, branches, repo_gid):
        filtered_branches = branches
        if self._branch and len(self._branch) > 0:
            filtered_branches = []
            input_branches = self._branch.split(",")
            if len(input_branches) == 1:
                if input_branches[0].lower() == "default".lower():
                    # Special case for default branch
                    self.connect()
                    repo_obj = self._get_repo_obj(repo_gid)
                    if repo_obj:
                        filtered_branches.append(repo_obj.default_branch)
                        return filtered_branches
            for b in branches:
                branch_name = b.name if hasattr(b, 'name') else b
                if branch_name in input_branches and branch_name not in filtered_branches:
                    filtered_branches.append(b)
        return filtered_branches

    def _get_branches(self, repo_gid, limit=None):
        branches = []
        if self._scan_scope:
            branches = self._scan_scope.get("repos", {}).get(repo_gid, {}).get("branches", {})
        if len(branches) > 0:
            return self._filter_branches(branches, repo_gid)
        self.connect()
        repo_obj = self._get_repo_obj(repo_gid)
        if repo_obj:
            branches = repo_obj.get_branches()
        return self._filter_branches(branches, repo_gid)

    def _get_files(self, repo_gid, branch_gid, limit=None):
        files = []
        files_content = []
        if self._scan_scope:
            files = self._scan_scope.get("repos", {}).get(repo_gid, {}).get("branches", {}).get(branch_gid, {}).get("files", [])
        if len(files) > 0:
            return files, files_content
        self.connect()
        repo_obj = self._get_repo_obj(repo_gid)
        if repo_obj:
            files = []
            try:
                contents = repo_obj.get_contents("", ref=branch_gid)
                while contents:
                    file_content = contents.pop(0)
                    if file_content.type == "dir":
                        contents.extend(repo_obj.get_contents(file_content.path, ref=branch_gid))
                    else:
                        files.append(file_content.path)
                        files_content.append(file_content)
            except Exception as e:
                message = f"Error listing files from branch {branch_gid}: {e}"
                self.log_message(message, logging.ERROR)
        return files, files_content

    def get_mapping(self, levels=-1, limit=None):
        if not self._client:
            return {}
        repos, owner = self._get_repos()
        map_data = {"owner": owner, "repos": {}}
        if repos:
            for repo in repos:
                repo_gid = repo.git_url
                message = f"Searching in repository: {repo.html_url}"
                self.log_message(message, logging.INFO)
                if len(repo_gid) > 0:
                    r_item = {
                        "gid": repo_gid,
                        "name": repo.name,
                        "branches": {}
                    }
                    map_data["repos"][repo_gid] = r_item
                    if levels > 0 and levels <= 1:
                        continue
                    if branches := self._get_branches(repo_gid, limit):
                        for branch in branches:
                            message = f"Searching in branch: {branch.name}"
                            self.log_message(message, logging.INFO)
                            branch_gid = branch.name
                            b_item = {
                                "gid": branch.commit.sha,
                                "name": branch.name,
                                "files": {}
                            }
                            if len(branch_gid) > 0:
                                map_data["repos"][repo_gid]["branches"][branch_gid] = b_item
                            if levels > 0 and levels <= 2:
                                continue
                            files, files_content = self._get_files(repo_gid, branch_gid)
                            map_data["repos"][repo_gid]["branches"][branch_gid]["files"] = files
                            if levels > 0 and levels <= 3:
                                continue
        return map_data

    def get_data(self, include_comments=False, limit=None):
        if not self._client:
            return {}

        repos = None
        q = self.get_query_from_scope()
        if q:
            repos = self._client.search_repositories(query=q)
        if not repos:
            repos, owner = self._get_repos()
        if repos:
            for repo in repos:
                if isinstance(repo, str):
                    repo = self._get_repo_obj(repo)
                repo_gid = repo.git_url
                repo_html_url = repo.html_url
                message = f"Searching in repository: {repo_html_url}"
                self.log_message(message, logging.INFO)

                # Iterate through each branch
                for branch in self._get_branches(repo_gid):
                    branch_gid = ""
                    if isinstance(branch, str):
                        branch_gid = branch
                    else:
                        branch_gid = branch.name
                    message = f"Searching in branch: {branch_gid}"
                    self.log_message(message, logging.INFO)

                    # Iterate through each file in the branch
                    try:
                        files, files_content = self._get_files(repo_gid, branch_gid)
                        use_preloaded_content = False
                        if len(files_content) > 0:
                            use_preloaded_content = True
                            files = files_content
                        for f in files:
                            try:
                                if use_preloaded_content:
                                    file_data = f.decoded_content.decode(errors='ignore')
                                    f = f.path
                                else:
                                    # Fetch file content
                                    file_data = repo.get_contents(f, ref=branch_gid).decoded_content.decode(errors='ignore')
                                url = repo.html_url + f"/blob/{branch_gid}/{f}"
                                file = self.pack_data(file_data, url)
                                yield file
                            except Exception as e:
                                message = f"Error accessing file {f} from branch {branch_gid}: {e}"
                                self.log_message(message, logging.ERROR)
                    except Exception as e:
                        message = f"Error accessing branch {branch_gid}: {e}"
                        self.log_message(message, logging.ERROR)

    def post_comment(self, issue, comment):
        if not self._client:
            return False
        message = f"Unable to post comment to {issue}!"
        self.log_message(message, logging.ERROR)
        return False

    def pack_data(self, file_data, url):
        ticket_data = {
            "ticket": {
                "file": {
                    "name": "file",
                    "data": file_data,
                    "data_type": "str"
                },
            },
            "url": url,
            "issue_id": url
        }
        return ticket_data



import logging


try:
    from . import hollow_controller as hollow_controller
except Exception:
    import n0s1.controllers.hollow_controller as hollow_controller


class AsanaController(hollow_controller.HollowController):
    def __init__(self):
        super().__init__()

    def set_config(self, config=None):
        super().set_config(config)
        import asana
        TOKEN = self._config.get("token", "")
        self._client = asana.Client.access_token(TOKEN)
        return self.is_connected()

    def get_name(self):
        return "Asana"

    def is_connected(self):
        if self._client:
            if user := self._client.users.get_user("me"):
                self.log_message(f"Logged to {self.get_name()} as {user}")
            else:
                self.log_message(f"Unable to connect to {self.get_name()} instance. Check your credentials.", logging.ERROR)
                return False

            if spaces := self._client.workspaces.get_workspaces():
                workspace_found = False
                project_found = False
                for s in spaces:
                    workspace_gid = s.get("gid", "")
                    if len(workspace_gid) > 0:
                        workspace_found = True
                        if projects := self._client.projects.get_projects_for_workspace(workspace_gid):
                            for p in projects:
                                project_found = True
                                break
                if workspace_found:
                    if project_found:
                        return True
                    else:
                        self.log_message(f"Unable to list {self.get_name()} projects. Check your permissions.", logging.ERROR)
                else:
                    self.log_message(f"Unable to list {self.get_name()} workspaces. Check your permissions.", logging.ERROR)
            else:
                self.log_message(f"Unable to connect to {self.get_name()} instance. Check your credentials.", logging.ERROR)
        return False

    def _get_workspaces(self, limit=None):
        workspaces = []
        if self._scan_scope:
            for key in self._scan_scope.get("workspaces", {}):
                w = self._client.workspaces.get_workspace(key)
                workspaces.append(w)
        if len(workspaces) > 0:
            return workspaces
        self.connect()
        return self._client.workspaces.get_workspaces()

    def _get_projects(self, workspace_gid, limit=None):
        projects = []
        if self._scan_scope:
            for key in self._scan_scope.get("workspaces", {}).get(workspace_gid, {}).get("projects", {}):
                p = self._client.projects.get_project(key)
                projects.append(p)
        if len(projects) > 0:
            return projects
        self.connect()
        return self._client.projects.get_projects_for_workspace(workspace_gid)

    def _get_tasks(self, workspace_gid, project_gid, limit=None):
        tasks = []
        if self._scan_scope:
            for key in self._scan_scope.get("workspaces", {}).get(workspace_gid, {}).get("projects", {}).get(project_gid, {}).get("tasks", {}):
                t = self._client.tasks.get_task(key, opt_fields=["name", "gid", "notes", "permalink_url"])
                tasks.append(t)
        if len(tasks) > 0:
            return tasks
        self.connect()
        return self._client.tasks.get_tasks_for_project(project_gid, opt_fields=["name", "gid", "notes", "permalink_url"])

    def _get_stories(self, workspace_gid, project_gid, task_gid, limit=None):
        stories = []
        if self._scan_scope:
            for key in self._scan_scope.get("workspaces", {}).get(workspace_gid, {}).get("projects", {}).get(project_gid, {}).get("tasks", {}).get(task_gid, {}).get("stories", {}):
                s = self._client.stories.get_story(key)
                stories.append(s)
        if stories:
            return stories
        self.connect()
        return self._client.stories.get_stories_for_task(task_gid)

    def get_mapping(self, levels=-1, limit=None):
        if not self._client:
            return {}
        map_data = {"workspaces": {}}
        if workspaces := self._get_workspaces(limit):
            for w in workspaces:
                workspace_gid = w.get("gid", "")
                if len(workspace_gid) > 0:
                    w_item = {
                        "gid": workspace_gid,
                        "name": w.get("name", ""),
                        "projects": {}
                    }
                    map_data["workspaces"][workspace_gid] = w_item
                    if levels > 0 and levels <= 1:
                        continue
                    if projects := self._get_projects(workspace_gid, limit):
                        for p in projects:
                            project_gid = p.get("gid", "")
                            p_item = {
                                "gid": project_gid,
                                "name": p.get("name", ""),
                                "tasks": {}
                            }
                            if len(project_gid) > 0:
                                map_data["workspaces"][workspace_gid]["projects"][project_gid] = p_item
                            if levels > 0 and levels <= 2:
                                continue
                            tasks = self._get_tasks(workspace_gid, project_gid)
                            for t in tasks:
                                task_gid = t.get("gid", "")
                                t_item = {
                                    "gid": task_gid,
                                    "name": t.get("name", ""),
                                    "stories": {}
                                }
                                if len(task_gid) > 0:
                                    map_data["workspaces"][workspace_gid]["projects"][project_gid]["tasks"][task_gid] = t_item
                                if levels > 0 and levels <= 3:
                                    continue
                                stories = self._get_stories(workspace_gid, project_gid, task_gid, limit)
                                for s in stories:
                                    story_gid = s.get("gid", "")
                                    s_item = {
                                        "gid": story_gid,
                                        "name": s.get("name", "")
                                    }
                                    if len(story_gid) > 0:
                                        map_data["workspaces"][workspace_gid]["projects"][project_gid]["tasks"][task_gid]["stories"][story_gid] = s_item

        return map_data

    def get_data(self, include_coments=False, limit=None):
        if not self._client:
            return {}

        if workspaces := self._get_workspaces(limit):
            for w in workspaces:
                workspace_gid = w.get("gid", "")
                if projects := self._get_projects(workspace_gid, limit):
                    for p in projects:
                        project_gid = p.get("gid", "")
                        if tasks := self._get_tasks(workspace_gid, project_gid, limit=None):
                            for t in tasks:
                                comments = []
                                title = t.get("name", "")
                                task_gid = t.get("gid", "")
                                description = t.get("notes", "")
                                url = t.get("permalink_url", "")
                                if include_coments:
                                    if stories := self._get_stories( workspace_gid, project_gid, task_gid, limit):
                                        for s in stories:
                                            if s.get("type", "").lower() == "comment".lower():
                                                comment = s.get("text", "")
                                                comments.append(comment)
                                ticket = self.pack_data(title, description, comments, url, task_gid)
                                yield ticket

    def post_comment(self, task_gid, comment):
        if not self._client:
            return False
        self.connect()
        if comment_status := self._client.stories.create_story_for_task(task_gid, {"type": "comment", "text": comment}):
            status = comment_status.get("text", "")
        return len(status) > 0



import logging
import time

try:
    from . import hollow_controller as hollow_controller
except Exception:
    import n0s1.controllers.hollow_controller as hollow_controller


class JiraController(hollow_controller.HollowController):
    def __init__(self):
        super().__init__()

    def set_config(self, config=None):
        super().set_config(config)
        from jira import JIRA
        SERVER = self._config.get("server", "")
        EMAIL = self._config.get("email", "")
        TOKEN = self._config.get("token", "")
        TIMEOUT = self._config.get("timeout", -1)
        VERIFY_SSL = not self._config.get("insecure", False)
        options = {"verify": VERIFY_SSL}
        if EMAIL and len(EMAIL) > 0:
            if TIMEOUT and TIMEOUT > 0:
                self._client = JIRA(SERVER, options=options, basic_auth=(EMAIL, TOKEN), timeout=TIMEOUT)
            else:
                self._client = JIRA(SERVER, options=options, basic_auth=(EMAIL, TOKEN))
        else:
            if TIMEOUT and TIMEOUT > 0:
                self._client = JIRA(SERVER, options=options, token_auth=TOKEN, timeout=TIMEOUT)
            else:
                self._client = JIRA(SERVER, options=options, token_auth=TOKEN)
        return self.is_connected()

    def get_name(self):
        return "Jira"

    def is_connected(self):
        from jira.exceptions import JIRAError
        try:
            if self._client:
                if user := self._client.myself():
                    self.log_message(f"Logged to {self.get_name()} as {user}")
                else:
                    self.log_message(f"Unable to connect to {self.get_name()} instance. Check your credentials.", logging.ERROR)
                    return False

                if projects := self._client.projects():
                    project_found = False
                    issue_found = False
                    for key in projects:
                        project_found = True
                        ql = f"project = '{key}'"
                        try:
                            for issue in self._client.search_issues(ql):
                                if issue:
                                    issue_found = True
                                    return True
                        except JIRAError as e:
                            if e.status_code == 400:
                                self.log_message(f"Skipping project '{key}' due to JIRAError 400: {e.text}",
                                                 logging.WARNING)
                                continue
                            else:
                                self.log_message(f"JIRAError: {e.status_code} {e.text}", logging.ERROR)
                                continue
                    if project_found:
                        if issue_found:
                            return True
                        else:
                            self.log_message(f"Unable to list {self.get_name()} issues. Check your permissions.", logging.ERROR)
                    else:
                        self.log_message(f"Unable to list {self.get_name()} projects. Check your permissions.", logging.ERROR)
                else:
                    self.log_message(f"Unable to connect to {self.get_name()} instance. Check your credentials.", logging.ERROR)
        except JIRAError as e:
            self.log_message(f"JIRAError: {e.status_code} {e.text}", logging.ERROR)
            self.log_message("Failed to retrieve user information. Check your credentials and permissions.",
                             logging.ERROR)
        return False

    def _get_projects(self, limit=None):
        if self._scan_scope:
            projects = []
            for key in self._scan_scope.get("projects", {}):
                projects.append(key)
            return projects
        self.connect()
        return self._client.projects()

    def _get_issues(self, project_key, limit=None):
        from jira.exceptions import JIRAError
        if not limit or limit < 0:
            limit = 50
        issue_keys = []
        if self._scan_scope:
            issue_keys = self._scan_scope.get("projects", {}).get(project_key, {})
        if len(issue_keys) > 0:
            counter = 0
            issues = []
            for key in issue_keys:
                counter += 1
                issue = self._client.issue(key)
                issues.append(issue)
                if counter > limit:
                    counter = 0
                    yield issues
                    issues = []
            if len(issues) > 0:
                yield issues
        else:
            ql = f"project = '{project_key}'"
            issues_finished = False
            nextPageToken = None
            while not issues_finished:
                try:
                    self.connect()
                    issues = self._client.enhanced_search_issues(ql, nextPageToken=nextPageToken, maxResults=limit)
                except JIRAError as e:
                    self.log_message(f"Error while searching issues on Jira project: [{project_key}]. Skipping...",
                                     logging.WARNING)
                    self.log_message(e)
                    issues = [{}]
                    break
                except Exception as e:
                    message = str(e) + f" client.enhanced_search_issues({ql}, nextPageToken={nextPageToken}, maxResults={limit})"
                    self.log_message(message, logging.WARNING)
                    issues = [{}]
                    time.sleep(1)
                    continue
                issues_finished = len(issues) <= 0
                nextPageToken = issues.nextPageToken
                if not nextPageToken:
                    issues_finished = True
                yield issues

    def get_mapping(self, levels=-1, limit=None):
        if not self._client:
            return {}
        map_data = {"projects": {}}
        if projects := self._get_projects(limit):
            for project in projects:
                map_data["projects"][str(project.key)] = {}
                if levels < 0 or levels > 1:
                    for issues in self._get_issues(str(project.key), limit):
                        for issue in issues:
                            map_data["projects"][str(project.key)][str(issue.key)] = {}
        return map_data

    def get_data(self, include_coments=False, limit=None):
        from jira.exceptions import JIRAError
        if not self._client:
            return {}
        try:
            self.connect()
            using_jql = False
            jql = self.get_query_from_scope()
            if jql:
                issues = self._client.search_issues(jql)
                for issue in issues:
                    ticket = self._extract_ticket(include_coments, issue)
                    using_jql = True
                    yield ticket
            if using_jql:
                projects = []
            else:
                projects = self._get_projects(limit)
        except Exception as e:
            message = str(e) + f" client.projects()"
            self.log_message(message, logging.WARNING)
            projects = []

        for key in projects:
            self.log_message(f"Scanning Jira project: [{key}]...")
            for issues in self._get_issues(key, limit):
                for issue in issues:
                    ticket = self._extract_ticket(include_coments, issue)
                    yield ticket

    def _extract_ticket(self, include_coments, issue):
        url = issue.self.split('/rest/api')[0] + "/browse/" + issue.key;
        title = issue.fields.summary
        description = issue.fields.description
        comments = []
        if include_coments:
            try:
                self.connect()
                issue_comments = self._client.comments(issue.id)
                comments.extend(c.body for c in issue_comments)
            except Exception as e:
                message = str(e) + f" client.comments({issue.id})"
                self.log_message(message, logging.WARNING)
                comments = []
                time.sleep(1)
        ticket = self.pack_data(title, description, comments, url, issue.key)
        return ticket

    def post_comment(self, issue, comment):
        if not self._client:
            return False
        comment = comment.replace("#", "0")
        self.connect()
        comment_status = self._client.add_comment(issue, body=comment)
        status = comment_status.id
        return bool(status and len(status) > 0 and int(status) > 0)



import logging


try:
    from . import hollow_controller as hollow_controller
except Exception:
    import n0s1.controllers.hollow_controller as hollow_controller


class GitLabController(hollow_controller.HollowController):
    def __init__(self):
        super().__init__()
        self._client = None

    def set_config(self, config=None):
        super().set_config(config)
        import gitlab
        TOKEN = config.get("token", "")
        URL = config.get("url", "https://gitlab.com")
        self._client = gitlab.Gitlab(URL, private_token=TOKEN)
        self._group = config.get("group", "")
        self._project = config.get("project", "")
        self._branch = config.get("branch", "")
        return self.is_connected()

    def get_name(self):
        return "GitLab"

    def is_connected(self):
        if self._client:
            try:
                self._client.auth()
                user = self._client.user
                self.log_message(f"Logged to {self.get_name()} as {user.username}")
                return True
            except Exception as e:
                self.log_message(f"Unable to connect to {self.get_name()}. Check your credentials: {e}", logging.ERROR)
                return False
        return False

    def _get_project_obj(self, project_id=None):
        if self._project and len(self._project) > 0:
            return self._client.projects.get(self._project)
        return self._client.projects.get(project_id)

    def _get_projects(self):
        owner = {}
        projects = []

        if self._scan_scope:
            owner = self._scan_scope.get("owner", {})
            for key in self._scan_scope.get("projects", {}):
                projects.append(key)
        if len(projects) > 0:
            return projects, owner
        
        self.connect()
        if self._group and len(self._group) > 0:
            try:
                group = self._client.groups.get(self._group)
                projects = group.projects.list(all=True)
                owner = {"type": "group", "name": self._group}
            except Exception as e:
                message = f"Unable to get projects from group {self._group}: {e}"
                self.log_message(message, logging.ERROR)
        else:
            projects = self._client.projects.list(all=True)
            owner = {"type": "authenticated_user", "name": self._client.user.username}

        if self._project and len(self._project) > 0:
            for p in projects:
                if p.path_with_namespace.lower() == self._project.lower() or str(p.id) == self._project:
                    return [p], owner
            try:
                # Try direct access by ID or path
                project = self._client.projects.get(self._project)
                return [project], owner
            except Exception:
                return [], owner

        return projects, owner

    def _filter_branches(self, branches, project_id):
        filtered_branches = branches
        if self._branch and len(self._branch) > 0:
            filtered_branches = []
            input_branches = self._branch.split(",")
            if len(input_branches) == 1:
                if input_branches[0].lower() == "default".lower():
                    # Special case for default branch
                    self.connect()
                    project_obj = self._get_project_obj(project_id)
                    if project_obj:
                        filtered_branches.append(project_obj.default_branch)
                        return filtered_branches
            for b in branches:
                branch_name = b.name if hasattr(b, 'name') else b
                if branch_name in input_branches and branch_name not in filtered_branches:
                    filtered_branches.append(b)
        return filtered_branches

    def _get_branches(self, project_id, limit=None):
        branches = []
        if self._scan_scope:
            project_id_str = str(project_id)
            branches = self._scan_scope.get("projects", {}).get(project_id_str, {}).get("branches", {})
        if len(branches) > 0:
            return self._filter_branches(branches, project_id)
        
        self.connect()
        project_obj = self._get_project_obj(project_id)
        if project_obj:
            branches = project_obj.branches.list(all=True)
        return self._filter_branches(branches, project_id)

    def _get_files(self, project_id, branch_name, path="", limit=None):
        files = []
        if self._scan_scope:
            files = self._scan_scope.get("projects", {}).get(project_id, {}).get("branches", {}).get(branch_name, {}).get("files", [])
        if len(files) > 0:
            return files
        
        self.connect()
        project_obj = self._get_project_obj(project_id)
        if project_obj:
            try:
                items = project_obj.repository_tree(path=path, ref=branch_name, recursive=True, all=True)
                for item in items:
                    if item['type'] == 'blob':
                        files.append(item['path'])
            except Exception as e:
                message = f"Error listing files from branch {branch_name}: {e}"
                self.log_message(message, logging.ERROR)
        return files

    def get_mapping(self, levels=-1, limit=None):
        if not self._client:
            return {}
        
        projects, owner = self._get_projects()
        map_data = {"owner": owner, "projects": {}}
        
        if projects:
            for project in projects:
                project_id = project.id if hasattr(project, 'id') else project
                project_name = project.path_with_namespace if hasattr(project, 'path_with_namespace') else project
                
                message = f"Searching in project: {project_name}"
                self.log_message(message, logging.INFO)
                
                if project_id:
                    p_item = {
                        "id": project_id,
                        "name": project_name,
                        "branches": {}
                    }
                    map_data["projects"][project_id] = p_item
                    
                    if levels > 0 and levels <= 1:
                        continue
                        
                    if branches := self._get_branches(project_id, limit):
                        for branch in branches:
                            branch_name = branch.name if hasattr(branch, 'name') else branch
                            message = f"Searching in branch: {branch_name}"
                            self.log_message(message, logging.INFO)
                            
                            b_item = {
                                "name": branch_name,
                                "files": {}
                            }
                            
                            if branch_name:
                                map_data["projects"][project_id]["branches"][branch_name] = b_item
                                
                            if levels > 0 and levels <= 2:
                                continue
                                
                            files = self._get_files(project_id, branch_name)
                            map_data["projects"][project_id]["branches"][branch_name]["files"] = files
                            
                            if levels > 0 and levels <= 3:
                                continue
        return map_data

    def get_data(self, include_comments=False, limit=None):
        if not self._client:
            return {}

        projects = None
        q = self.get_query_from_scope()
        if q:
            projects = self._client.projects.list(search=q, all=True)
        if not projects:
            projects, owner = self._get_projects()
            
        if projects:
            for project in projects:
                # Always get the full project object to ensure all attributes are available
                if hasattr(project, 'id'):
                    project_id = project.id
                    # Get the full project object
                    try:
                        project = self._client.projects.get(project_id)
                    except Exception as e:
                        message = f"Error getting full project object for ID {project_id}: {e}"
                        self.log_message(message, logging.ERROR)
                        continue
                else:
                    project_id = project
                    try:
                        project = self._get_project_obj(project_id)
                    except Exception as e:
                        message = f"Error getting project object for ID {project_id}: {e}"
                        self.log_message(message, logging.ERROR)
                        continue
                
                project_name = project.path_with_namespace
                project_url = project.web_url
                project_id = project.id
                
                message = f"Searching in project: {project_name}"
                self.log_message(message, logging.INFO)

                # Iterate through each branch
                for branch in self._get_branches(project_id):
                    branch_name = branch.name if hasattr(branch, 'name') else branch
                    message = f"Searching in branch: {branch_name}"
                    self.log_message(message, logging.INFO)

                    # Iterate through each file in the branch
                    try:
                        files = self._get_files(project_id, branch_name)
                        for file_path in files:
                            try:
                                # Fetch file content
                                file_content = project.files.get(file_path=file_path, ref=branch_name)
                                raw_content = file_content.decode()
                                # Properly decode bytes to string
                                if isinstance(raw_content, bytes):
                                    file_data = raw_content.decode('utf-8', errors='replace')
                                else:
                                    file_data = raw_content
                                url = f"{project_url}/-/blob/{branch_name}/{file_path}"
                                file = self.pack_data(file_data, url)
                                yield file
                            except Exception as e:
                                message = f"Error accessing file {file_path} from branch {branch_name}: {e}"
                                self.log_message(message, logging.ERROR)
                    except Exception as e:
                        message = f"Error accessing branch {branch_name}: {e}"
                        self.log_message(message, logging.ERROR)

    def post_comment(self, issue, comment):
        if not self._client:
            return False
        message = f"Unable to post comment to {issue}!"
        self.log_message(message, logging.ERROR)
        return False

    def pack_data(self, file_data, url):
        ticket_data = {
            "ticket": {
                "file": {
                    "name": "file",
                    "data": file_data,
                    "data_type": "str"
                },
            },
            "url": url,
            "issue_id": url
        }
        return ticket_data


import hashlib


def sha1_hash(to_hash):
    try:
        message_digest = hashlib.sha256()
        string_m = str(to_hash)
        byte_m = bytes(string_m, encoding='utf')
        message_digest.update(byte_m)
        return message_digest.hexdigest()
    except TypeError as e:
        raise "Unable to generate SHA-256 hash for input string" from e


def _safe_re_search(regex_str, text):
    global DEBUG
    m = None
    try:
        m = re.search(regex_str, text)
    except Exception:
        try:
            regex_str = regex_str.replace("(?i)", "")
            m = re.search(regex_str, text, re.IGNORECASE)
        except Exception as e:
            if DEBUG:
                log_message(str(e))
    return m

def sanitize_text(text, begin, end):
    text_len = len(text)
    s_begin = max(begin - 20, 0)
    s_end = min(end + 20, text_len)
    sanitized_text = f"{text[s_begin:begin]}<REDACTED>{text[end:s_end]}"
    snippet_text = text[s_begin:s_end]
    return sanitized_text, snippet_text


def match_regex(regex_config, text):
    for c in regex_config["rules"]:
        regex_str = c["regex"]
        if m := _safe_re_search(regex_str, text):
            begin = m.regs[0][0]
            end = m.regs[0][1]
            matched_text = text[begin:end]
            sanitized_text, snippet_text = sanitize_text(text, begin, end)
            return c, matched_text, sanitized_text, snippet_text
    return None, None, None, None


# def callback(a, b):
#     print('Sum = {0}'.format(a+b))
#
# def main(a,b,f=None):
#     print('Add any two digits.')
#     if f is not None:
#         f(a,b)
#
# main(1, 2, callback)

# def is_connected(self):
#     try:
#         if self._client:
#             if user := self._client.myself():
#                 self.log_message(f"Logged to {self.get_name()} as {user}")
#             else:
#                 self.log_message(f"Unable to connect to {self.get_name()} instance. Check your credentials.", logging.ERROR)
#                 return False
#
#         if projects := self._client.projects():
#             project_found = False
#             issue_found = False
#             for key in projects:
#                 project_found = True
#                 ql = f"project = '{key}'"
#                 try:
#                     for issue in self._client.search_issues(ql):
#                         if issue:
#                             issue_found = True
#                             return True
#                 except JIRAError as e:
#                     if e.status_code == 400:
#                         self.log_message(f"Skipping project '{key}' due to JIRAError 400: {e.text}", logging.WARNING)
#                         continue
#                     else:
#                         self.log_message(f"JIRAError: {e.status_code} {e.text}", logging.ERROR)
#                         continue
#                 if project_found:
#                     if issue_found: return True
#                     else: self.log_message(f"Unable to list {self.get_name()} issues. Check your permissions.", logging.ERROR)
#                 else:
#                     self.log_message(f"Unable to list {self.get_name()} projects. Check your permissions.", logging.ERROR)
#             else:
#                 self.log_message(f"Unable to connect to {self.get_name()} instance. Check your credentials.", logging.ERROR)
#                 return False
#     except JIRAError as e:
#         self.log_message(f"JIRAError: {e.status_code} {e.text}", logging.ERROR)
#         self.log_message("Failed to retrieve user information. Check your credentials and permissions.", logging.ERROR)
#     return False
#
# Console Output INFO:root:Logged to Jira as {'self': 'https://<JIRA_INSTANCE>/rest/api/2/user?accountId=<ACCOUNT_ID>', 'accountId': '<ACCOUNT_ID>', 'accountType': 'atlassian', 'emailAddress': '<EMAIL_ADDRESS>', 'avatarUrls': {'48x48': '<AVATAR_URL_48>', '24x24': '<AVATAR_URL_24>', '16x16': '<AVATAR_URL_16>', '32x32': '<AVATAR_URL_32>'}, 'displayName': '<DISPLAY_NAME>', 'active': True, 'timeZone': '<TIME_ZONE>', 'locale': '<LOCALE>', 'groups': {'size': 5, 'items': []}, 'applicationRoles': {'size': 1, 'items': []}, 'expand': 'groups,applicationRoles'} WARNING:root:Skipping project '<PROJECT_KEY_1>' due to JIRAError 400: The value '<PROJECT_KEY_1>' does not exist for the field 'project'. WARNING:root:Skipping project '<PROJECT_KEY_2>' due to JIRAError 400: The value '<PROJECT_KEY_2>' does not exist for the field 'project'. INFO:root:Starting scan in community mode... INFO:root:Scanning Jira project: [TEST]...
#

from langchain_text_splitters import RecursiveJsonSplitter

json_data = {
    "complex_key": {
        "nested_key": [
            {"item1": "value1"},
            {"item2": "value2"},
            {"item3": "value3"}
        ]
    },
    "another_key": "some_value"
}

splitter = RecursiveJsonSplitter(max_chunk_size=5)
json_chunks = splitter.split_json(json_data)

print(json_chunks)


import html
import logging
import time


try:
    from . import hollow_controller as hollow_controller
except Exception:
    import n0s1.controllers.hollow_controller as hollow_controller


class ConfluenceController(hollow_controller.HollowController):
    def __init__(self):
        super().__init__()
        self._url = None
        self._user = None
        self._password = None

    def set_config(self, config=None):
        super().set_config(config)
        from atlassian import Confluence
        SERVER = self._config.get("server", "")
        EMAIL = self._config.get("email", "")
        TOKEN = self._config.get("token", "")
        TIMEOUT = self._config.get("timeout", -1)
        VERIFY_SSL = not self._config.get("insecure", False)
        self._url = SERVER
        self._user = EMAIL
        self._password = TOKEN
        if EMAIL and len(EMAIL) > 0:
            if TIMEOUT and TIMEOUT > 0:
                self._client = Confluence(url=SERVER, verify_ssl=VERIFY_SSL, username=EMAIL, password=TOKEN, timeout=TIMEOUT)
            else:
                self._client = Confluence(url=SERVER, verify_ssl=VERIFY_SSL, username=EMAIL, password=TOKEN)
        else:
            if TIMEOUT and TIMEOUT > 0:
                self._client = Confluence(url=SERVER, verify_ssl=VERIFY_SSL, token=TOKEN, timeout=TIMEOUT)
            else:
                self._client = Confluence(url=SERVER, verify_ssl=VERIFY_SSL, token=TOKEN)
        return self.is_connected()

    def get_name(self):
        return "Confluence"

    def _get_request(self, url):
        from requests.auth import HTTPBasicAuth
        import requests
        response = None
        try:
            if self._user:
                auth = HTTPBasicAuth(self._user, self._password)
                headers = {
                    "Content-Type": "application/json"
                }
                response = requests.request(
                    "GET",
                    url,
                    headers=headers,
                    auth=auth
                )
            else:
                headers = {
                    "Content-Type": "application/json",
                    "Authorization": f"Bearer {self._password}"
                }
                response = requests.request(
                    "GET",
                    url,
                    headers=headers
                )
        except Exception as e:
            self.log_message(str(e))
        return response

    def get_current_user(self):
        url = f"{self._url}/rest/api/user/current"
        response = self._get_request(url)
        user = response.json() if response and response.status_code == 200 else None
        if user:
            user_type = user.get("type", "")
            if len(user_type) > 0:
                return user
        else:
            url = f"{self._url}/wiki/rest/api/user/current"
            response = self._get_request(url)
            if response and response.status_code == 200:
                user = response.json()
            if user:
                user_type = user.get("type", "")
                if len(user_type) > 0:
                    return user
        return None

    def is_connected(self):
        if self._client:
            if user := self.get_current_user():
                self.log_message(f"Logged to {self.get_name()} as {user}")
            else:
                self.log_message(f"Unable to connect to {self.get_name()} instance. Check your credentials.", logging.ERROR)
                return False

            if spaces := self._client.get_all_spaces():
                space_found = False
                page_found = False
                for s in spaces.get("results", []):
                    key = s.get("key", "")
                    if len(key) > 0:
                        space_found = True
                        pages = self._client.get_all_pages_from_space(key)
                        if len(pages) > 0:
                            page_found = True
                            break
                if space_found:
                    if page_found:
                        return True
                    else:
                        self.log_message(f"Unable to list {self.get_name()} pages. Check your permissions.", logging.ERROR)
                else:
                    self.log_message(f"Unable to list {self.get_name()} spaces. Check your permissions.", logging.ERROR)
            else:
                self.log_message(f"Unable to connect to {self.get_name()} instance. Check your credentials.", logging.ERROR)
        return False

    def _get_workspaces(self, limit=None):
        if self._scan_scope:
            workspaces = []
            for key in self._scan_scope.get("workspaces", {}):
                workspaces.append(key)
            yield workspaces
        else:
            space_start = 0
            if not limit or limit < 0:
                limit = 50
            finished = False
            while not finished:
                try:
                    self.connect()
                    res = self._client.get_all_spaces(start=space_start, limit=limit)
                    spaces = res.get("results", [])
                except Exception as e:
                    message = str(e) + f" get_all_spaces(start={space_start}, limit={limit})"
                    self.log_message(message, logging.WARNING)
                    spaces = [{}]
                    time.sleep(1)
                    continue
                space_start += limit
                if len(spaces) <= 0:
                    finished = True
                yield spaces

    def _get_pages(self, workspace_key, limit=None):
        from atlassian.confluence import ApiPermissionError
        start = 0
        if not limit or limit < 0:
            limit = 50
        pages_start = start
        page_keys = []
        if self._scan_scope:
            page_keys = self._scan_scope.get("workspaces", {}).get(workspace_key, {})
        if len(page_keys) > 0:
            counter = 0
            pages = []
            for key in page_keys:
                counter += 1
                page = self._client.get_page_by_id(key)
                pages.append(page)
                if counter > limit:
                    counter = 0
                    yield pages
                    pages = []
            if len(pages) > 0:
                yield pages
        else:
            if len(workspace_key) > 0:
                pages_finished = False
                while not pages_finished:
                    try:
                        self.connect()
                        pages = self._client.get_all_pages_from_space(workspace_key, start=pages_start, limit=limit)
                    except ApiPermissionError as e:
                        message = str(e) + f" get_all_pages_from_space({workspace_key}, start={pages_start}, limit={limit}). Skipping..."
                        self.log_message(message, logging.WARNING)
                        pages = [{}]
                        break
                    except Exception as e:
                        message = str(e) + f" get_all_pages_from_space({workspace_key}, start={pages_start}, limit={limit})"
                        self.log_message(message, logging.WARNING)
                        pages = [{}]
                        time.sleep(1)
                        continue
                    pages_start += limit
                    if len(pages) <= 0:
                        pages_finished = True
                    yield pages

    def get_mapping(self, levels=-1, limit=None):
        if not self._client:
            return {}
        map_data = {"workspaces": {}}
        for spaces in self._get_workspaces(limit):
            for space in spaces:
                workspace_key = space.get("key", None)
                if workspace_key:
                    map_data["workspaces"][workspace_key] = {}
                    if levels < 0 or levels > 1:
                        for pages in self._get_pages(workspace_key, limit):
                            for page in pages:
                                page_id = page.get("id", None)
                                page_title = page.get("title", "")
                                if page_id:
                                    map_data["workspaces"][workspace_key][page_id] = {"title": page_title}
        return map_data

    def get_data(self, include_coments=False, limit=None):
        if not self._client:
            return {}

        pages = []
        using_cql = False
        cql = self.get_query_from_scope()
        if cql:
            try:
                res = self._client.cql(cql, limit=limit)
                results = res.get("results", [])
                for r in results:
                    content_type = r.get("content", {}).get("type", None)
                    if content_type and content_type.lower() == "page".lower():
                        pages.append(r.get("content", {}))
                if len(pages) > 0:
                    using_cql = True
                    yield from self.process_pages(include_coments, limit, pages)
                else:
                    message = f"No pages found for [cql:{cql}]. Scan will not be scoped."
                    self.log_message(message, logging.WARNING)
                    self._scan_scope = None
            except Exception as e:
                message = str(e) + f" cql({cql}, limit={limit})"
                self.log_message(message, logging.WARNING)

        if using_cql:
            return

        for spaces in self._get_workspaces(limit):
            for s in spaces:
                key = s
                if isinstance(s, dict):
                    key = s.get("key", "")
                self.log_message(f"Scanning Confluence space: [{key}]...")
                if len(key) > 0:
                    for pages in self._get_pages(key, limit):
                        yield from self.process_pages(include_coments, limit, pages)

    def process_pages(self, include_coments, limit, pages):
        for p in pages:
            comments = []
            title = p.get("title", "")
            page_id = p.get("id", "")
            try:
                self.connect()
                body = self._client.get_page_by_id(page_id, expand="body.storage")
            except Exception as e:
                message = str(e) + f" get_page_by_id({page_id})"
                self.log_message(message, logging.WARNING)
                body = {}
                time.sleep(1)
                continue

            description = body.get("body", {}).get("storage", {}).get("value", "")
            url = body.get("_links", {}).get("base", "") + p.get("_links", {}).get("webui", "")
            if len(page_id) > 0 and include_coments:
                if not limit or limit < 0:
                    limit = 50
                comments_start = 0
                comments_finished = False
                while not comments_finished:
                    try:
                        self.connect()
                        comments_response = self._client.get_page_comments(page_id, expand="body.storage",
                                                                           start=comments_start, limit=limit)
                        comments_result = comments_response.get("results", [])
                    except Exception as e:
                        message = str(
                            e) + f" get_page_comments({page_id}, expand=\"body.storage\", start={comments_start}, limit={limit})"
                        self.log_message(message, logging.WARNING)
                        comments_result = [{}]
                        time.sleep(1)
                        continue
                    comments_start += limit

                    for c in comments_result:
                        comment = c.get("body", {}).get("storage", {}).get("value", "")
                        comments.append(comment)

                    if len(comments_result) <= 0:
                        comments_finished = True

            ticket = self.pack_data(title, description, comments, url, page_id)
            yield ticket

    def post_comment(self, issue, comment):
        if not self._client:
            return False
        comment = comment.replace("#", "0")
        comment = html.escape(comment, quote=True)
        status = -1
        self.connect()
        if comment_status := self._client.add_comment(issue, comment):
            status = comment_status.get("id", "")
        return int(status) > 0



import logging


class HollowController:
    def __init__(self):
        self._client = None
        self._config = None
        self._check_connection_after = 200
        self._requests_counter = 0
        self._scan_scope = None
        self.log_message_callback = None

    def set_config(self, config=None):
        if config:
            self._config = config
            self._scan_scope = self._config.get("scan_scope", None)
        return self._config is not None

    def set_log_message_callback(self, log_message_callback):
        self.log_message_callback = log_message_callback

    def get_name(self):
        return "Hollow"

    def connect(self):
        self._requests_counter += 1
        if self._requests_counter > self._check_connection_after:
            self._requests_counter = 0
            if not self.is_connected():
                # Force new connection
                self._client = None

        if not self._client:
            if self.set_config():
                return self.is_connected()
        return True

    def is_connected(self):
        return False

    def get_data(self, include_coments=False, limit=None):
        return {}

    def get_mapping(self, levels=-1, limit=None):
        return {}

    def post_comment(self, issue, comment):
        return self.is_connected()

    def log_message(self, message, level=logging.INFO):
        if self.log_message_callback:
            self.log_message_callback(message, level)
        else:
            print(message)

    def pack_data(self, title, description, comments, url, ticket_key):
        ticket_data = {
            "ticket": {
                "title": {
                    "name": "title",
                    "data": title,
                    "data_type": "str"
                },
                "description": {
                    "name": "description",
                    "data": description,
                    "data_type": "str"
                },
                "comments": {
                    "name": "comments",
                    "data": comments,
                    "data_type": "list"
                }
            },
            "url": url,
            "issue_id": ticket_key
        }
        return ticket_data

    def get_query_from_scope(self):
        query = None
        if self._scan_scope:
            query = self._scan_scope.get("query", None)
            if not query:
                query = self._scan_scope.get("search", None)
            if not query:
                query = self._scan_scope.get("jql", None)
            if not query:
                query = self._scan_scope.get("cql", None)
        return query



import logging


try:
    from . import hollow_controller as hollow_controller
except Exception:
    import n0s1.controllers.hollow_controller as hollow_controller


class ZendeskController(hollow_controller.HollowController):
    def __init__(self):
        super().__init__()
        self._client = None

    def set_config(self, config):
        super().set_config(config)
        from zenpy import Zenpy
        SERVER = self._config.get("server", "")
        EMAIL = self._config.get("email", "")
        TOKEN = self._config.get("token", "")
        creds = {
            "email": EMAIL,
            "token": TOKEN,
            "subdomain": SERVER
        }
        self._client = Zenpy(**creds)
        return self.is_connected()

    def get_name(self):
        return "Zendesk"

    def is_connected(self):
        if self._client:
            if user := self._client.users.me():
                self.log_message(f"Logged to {self.get_name()} as {user} - {user.email}")
                return True
            else:
                self.log_message(f"Unable to connect to {self.get_name()}. Check your credentials.", logging.ERROR)
                return False
        return False

    def get_data(self, include_comments=False, limit=None):
        if not self._client:
            return {}

        try:
            server = self._config.get("server", "")

            using_scan_scope = False
            query = self.get_query_from_scope()
            if query:
                tickets = self._client.search(query=query)
                if len(tickets) > 0:
                    using_scan_scope = True

            if not using_scan_scope:
                # Fetch all tickets (paginated)
                tickets = self._client.tickets()

            for ticket in tickets:
                self.log_message(f"Scanning Zendesk Ticket ID: {ticket.id}, Subject: {ticket.subject}, Status: {ticket.status}, Created: {ticket.created_at}")
                comments = []
                title = ticket.subject
                ticket_id = ticket.id
                description = ticket.description
                url = ticket.url
                if len(server):
                    url = f"https://{server}.zendesk.com/agent/tickets/{ticket_id}"
                if include_comments:
                    if cs := self._client.tickets.comments(ticket_id):
                        for c in cs:
                            c_body = c.body
                            comments.append(c_body)
                ticket = self.pack_data(title, description, comments, url, ticket_id)
                yield ticket
        except Exception as e:
            message = str(e) + f" client.get_data()"
            self.log_message(message, logging.WARNING)

    def post_comment(self, issue, comment):
        if not self._client:
            return False
        from zenpy.lib.api_objects import Ticket, Comment
        c = Comment(body=comment, public=True)
        ticket = Ticket(id=issue, comment=c)
        return self._client.tickets.update(ticket)



import secret_scan
import os

debug=False
email="marcelo@spark1.us"
api_key = os.getenv("JIRA_TOKEN")
server="https://spark1us.atlassian.net"
scope = {"cql": "space=IS and type=page"}
scope=None

# result = secret_scan.confluence_scan(debug=debug, email=email, server=server, api_key=api_key, scan_scope=scope)
# print(result)

result = secret_scan.jira_scan(debug=debug, email=email, server=server, api_key=api_key)
print(result)